#! /usr/bin/env python

import sys
import os
import tempfile


class SingleInstanceException(BaseException):
    pass


class SingleInstance:

    """
    If you want to prevent your script from running in parallel just instantiate SingleInstance() class.
    If is there another instance already running it will throw a `SingleInstanceException`.
    >>> import tendo
    ... me = SingleInstance()
    This option is very useful if you have scripts executed by crontab at small amounts of time.
    Remember that this works by creating a lock file with a filename based on the full path to the script file.
    Providing a flavor_id will augment the filename with the provided flavor_id, allowing you to create
    multiple singleton instances from the same file. This is particularly useful if you want specific
    functions to have their own singleton instances.
    """

    def __init__(self, flavor_id=""):
        import sys
        self.initialized = False
        basename = os.path.splitext(os.path.abspath(sys.argv[0]))[0].replace(
            "/", "-").replace(":", "").replace("\\", "-") + '-%s' % flavor_id + '.lock'
        # os.path.splitext(os.path.abspath(sys.modules['__main__'].__file__))[0].replace("/", "-").
        # replace(":", "").replace("\\", "-") + '-%s' % flavor_id + '.lock'
        self.lockfile = os.path.normpath(
            tempfile.gettempdir() + '/' + basename)

        if sys.platform == 'win32':
            try:
                # file already exists, we try to remove (in case previous
                # execution was interrupted)
                if os.path.exists(self.lockfile):
                    os.unlink(self.lockfile)
                self.fd = os.open(
                    self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            except OSError:
                type, e, tb = sys.exc_info()
                if e.errno == 13:
                    return
        else:  # non Windows
            import fcntl
            self.fp = open(self.lockfile, 'w')
            self.fp.flush()
            try:
                fcntl.lockf(self.fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError:
                return
        self.initialized = True

    def __del__(self):
        import sys
        import os
        if not self.initialized:
            return
        try:
            if sys.platform == 'win32':
                if hasattr(self, 'fd'):
                    os.close(self.fd)
                    os.unlink(self.lockfile)
            else:
                import fcntl
                fcntl.lockf(self.fp, fcntl.LOCK_UN)
                # os.close(self.fp)
                if os.path.isfile(self.lockfile):
                    os.unlink(self.lockfile)
        except Exception as e:
            raise


def f(name):
    # tmp = logger.level
    # logger.setLevel(logging.CRITICAL)  # we do not want to see the warning
    try:
        me2 = SingleInstance(flavor_id=name)  # noqa
    except SingleInstanceException:
        sys.exit(-1)
    # logger.setLevel(tmp)
    pass



import unittest
from tempfile import mkstemp
from os import remove
from datetime import date, datetime, timedelta
from sys import version

from qtodotxt2.lib.file import File
from qtodotxt2.lib.tasklib import Task
from qtodotxt2.lib.filters import DueTodayFilter, DueTomorrowFilter, DueThisWeekFilter, DueThisMonthFilter, DueOverdueFilter


PYTHON_VERSION = version[:3]


if PYTHON_VERSION < '3.3':
    FileNotFoundError = BaseException


class TestFile(unittest.TestCase):
    def setUp(self):
        self.file = File()
        self.tmpfile = mkstemp(text=True)[1]

    def tearDown(self):
        try:
            remove(self.tmpfile)
        except FileNotFoundError:
            pass
        except OSError as ex:    # maintain compatibility with Python 3.2
            if ex.errno != 2:
                raise
        except:
            raise

    def saveAndReload(self):
        self.file.save(self.tmpfile)
        self.file = File()
        self.file.load(self.tmpfile)

    def test_single_task(self):
        text = 'due:1999-10-10 do something +project1 @context1'
        self.file.tasks.append(Task(text))
        self.saveAndReload()
        self.assertEqual(self.file.tasks[0].text, text)
        self.assertEqual(self.file.tasks[0].contexts, ['context1'])
        self.assertEqual(self.file.tasks[0].projects, ['project1'])
        self.assertFalse(self.file.tasks[0].is_complete)
        self.assertFalse(self.file.tasks[0].priority)
        self.assertEqual(self.file.tasks[0].due, datetime(1999, 10, 10))

    def test_two_tasks(self):
        task1 = 'do something +project1 @context1'
        task2 = '(A) do something else +project1 @context2'
        self.file.tasks.extend([
            Task(task1),
            Task(task2)
        ])
        self.saveAndReload()
        self.assertEqual(self.file.tasks[0].text, task1)
        self.assertEqual(self.file.tasks[1].text, task2)

    def test_five_tasks(self):
        task1 = Task('do something +project1 @context1')
        task2 = Task('(A) do something else +project1 @context2')
        task3 = Task('do something else +project1 @context2')
        task4 = Task('something else +project1 @context2')
        task5 = Task('abc +project1 @context2')
        self.file.tasks.extend([task1, task2, task3, task4, task5])
        self.saveAndReload()
        self.assertEqual(self.file.tasks[0].text, task1.text)
        self.assertEqual(self.file.tasks[1].text, task2.text)
        self.assertEqual(self.file.tasks[2].text, task3.text)
        self.assertEqual(self.file.tasks[3].text, task4.text)
        self.assertEqual(self.file.tasks[4].text, task5.text)

    def test_get_all_contexts(self):
        self.file.tasks.extend([
            Task('x task with @context1'),
            Task('task with @context2'),
            Task('task with @context1 and @context2'),
            Task('task with @context1 and @context2 and @context3')
        ])
        self.saveAndReload()
        self.assertEqual(self.file.getAllContexts(), {'context1': [2, 1], 'context2': [3, 0], 'context3': [1, 0]})

    def test_get_all_incl_completed_contexts(self):
        self.file.tasks.extend([
            Task('x task with @context1'),
            Task('task with @context2'),
            Task('task with @context1 and @context2'),
            Task('x task with @context1 and @context2 and @context3')
        ])
        self.saveAndReload()
        self.assertEqual(self.file.getAllContexts(), {'context1': [1, 2], 'context2': [2, 1], 'context3': [0, 1]})

    def test_get_all_projects(self):
        self.file.tasks.extend([
            Task('x task with +project1'),
            Task('task with +project2'),
            Task('task with +project1 and +project2'),
            Task('task with +project1 and +project2 and +project3')
        ])
        self.saveAndReload()
        self.assertEqual(self.file.getAllProjects(), {'project1': [2, 1], 'project2': [3, 0], 'project3': [1, 0]})

    def test_get_all_due_ranges(self):
        today = date.today().strftime('%Y-%m-%d')
        yesterday = (date.today() - timedelta(days=1)).strftime('%Y-%m-%d')

        self.file.tasks.extend([
            Task('x due:' + today + ' completed task of today'),
            Task('due:' + today + ' first task of today'),
            Task('due:' + today + ' second task of today'),
            Task('due:' + yesterday + ' task of yesterday'),
        ])
        self.saveAndReload()
        due = self.file.getAllDueRanges()
        print(due)
        self.assertEqual(due[DueTodayFilter()], [2, 1])
        self.assertEqual(due[DueOverdueFilter()], [1, 0])
        self.assertEqual(due[DueThisWeekFilter()], [2, 1])

    def test_load_empty_filename(self):
        self.assertRaises(OSError, self.file.load, '')

    def test_load_nonexisting_file(self):
        remove(self.tmpfile)
        self.assertRaises(OSError, self.file.load, self.tmpfile)



from datetime import date
import unittest
from qtodotxt2.lib.tasklib import Task, RecursiveMode


class TestTasks(unittest.TestCase):

    def test_completeness_comparison(self):
        self.assertEqual(Task('task1').is_complete, Task('task2').is_complete)
        self.assertEqual(Task('x task1').is_complete, Task('x task2').is_complete)
        self.assertNotEqual(Task('task').is_complete, Task('x task').is_complete)
        self.assertNotEqual(Task('x task').is_complete, Task('task').is_complete)
        self.assertLess(Task('task'), Task('x task'))
        self.assertGreater(Task('x task'), Task('task'))

    def test_priority_comparison(self):
        # this tests are a bit broken now that priority is only a character
        self.assertEqual(Task('task1').priority, Task('task2').priority)
        self.assertEqual(Task('(A) task1').priority, Task('(A) task2').priority)
        self.assertNotEqual(Task('(A) task').priority, Task('task').priority)
        self.assertLess(Task('(A) task').priority, Task('(B) task').priority)
        self.assertGreater(Task('(B) task').priority, Task('(A) task').priority)
        self.assertNotEqual(Task('(A) task1').priority, Task('(AA) task2').priority)
        self.assertEqual(Task('(1) task1').priority, Task('(1) task2').priority)

    def test_comparisons_comparison(self):
        self.assertEqual(Task('task'), Task('task'))
        self.assertEqual(Task('(A) task'), Task('(A) task'))

        # self.assertGreater(Task('task1'), Task('task2'))
        self.assertGreater(Task('x task'), Task('task'))
        self.assertLess(Task('(A) task'), Task('(B) task'))
        self.assertLess(Task('(A) task'), Task('task'))
        self.assertLess(Task('(A) task'), Task('x (A) task'))

        # self.assertLess(Task('task2'), Task('task1'))
        self.assertGreater(Task('x task'), Task('task'))
        self.assertGreater(Task('(B) task'), Task('(A) task'))
        self.assertGreater(Task('task'), Task('(A) task'))
        self.assertGreater(Task('x (A) task'), Task('(A) task'))

    def test_task_ordering(self):
        task1 = Task('x task1')
        task2 = Task('task2')
        task3 = Task('abc task2')
        task4 = Task('(D) task2')
        task5 = Task('(C) b task')
        task6 = Task('(C) a task')
        task7 = Task('(B) task2')
        task8 = Task('(A) task2')
        self.assertTrue(task2 > task3)
        self.assertTrue(task1 > task2 > task3 > task4 > task5 > task6 > task7 > task8)

    def test_priority(self):
        self.assertEqual(Task("task").priority, "")
        self.assertEqual(Task("(a) task").priority, "")
        self.assertEqual(Task("x (a) task").priority, "")
        self.assertEqual(Task("x 2016-03-12 task").priority, "")
        self.assertEqual(Task("(A) task").priority, "A")
        t = Task("(A) task")
        self.assertEqual(t.priority, "A")
        t.increasePriority()
        self.assertEqual(t.priority, "A")
        t.decreasePriority()
        self.assertEqual(t.priority, "B")
        t.decreasePriority()
        t.decreasePriority()
        self.assertEqual(t.priority, "D")
        self.assertEqual(t.text, "(D) task")
        t.decreasePriority()
        self.assertEqual(t.priority, "")
        self.assertEqual(t.text, "task")

        t.increasePriority()
        t.increasePriority()
        self.assertEqual(t.priority, "C")
        for i in range(20):
            t.increasePriority()
        self.assertEqual(t.priority, "A")
        for i in range(20):
            t.decreasePriority()
        self.assertEqual(t.priority, "")

        # this task i wrongly formated, x should be followed by adate
        # self.assertEqual(Task("x (A) task").priority, Priority("A"))

        # A task with a priority lower than our default minimal priority
        t = Task("(M) task")
        t.increasePriority()
        self.assertEqual(t.priority, "L")
        t.decreasePriority()
        self.assertEqual(t.priority, "")

    def test_basic(self):
        task = Task('do something')
        self.assertEqual(task.text, 'do something')
        self.assertEqual(len(task.contexts), 0)
        self.assertFalse(len(task.projects), 0)
        self.assertFalse(task.is_complete)
        self.assertFalse(task.priority)

        task = Task('do something @context1 @context2')
        self.assertEqual(task.contexts, ['context1', 'context2'])
        self.assertEqual(task.projects, [])
        self.assertFalse(task.is_complete)
        self.assertFalse(task.priority)

        task = Task('do something +project1 +project2')
        self.assertEqual(task.contexts, [])
        self.assertEqual(task.projects, ['project1', 'project2'])
        self.assertFalse(task.is_complete)
        self.assertFalse(task.priority)

        task = Task('(E) do something +project1 @context1 +project2 rest of line @context2')
        self.assertEqual(task.contexts, ['context1', 'context2'])
        self.assertEqual(task.projects, ['project1', 'project2'])
        self.assertFalse(task.is_complete)
        self.assertTrue(task.priority)
        self.assertEqual(task.text, '(E) do something +project1 @context1 +project2 rest of line @context2')

        # task with + alone and complete
        task = Task('x 2016-01-23 do something +project1 @context1 +project2 rest + of line @context2')
        self.assertEqual(task.contexts, ['context1', 'context2'])
        self.assertEqual(task.projects, ['project1', 'project2'])
        self.assertTrue(task.is_complete)
        self.assertFalse(task.priority)
        self.assertEqual(task.completion_date, date(2016, 1, 23))

    def test_completion(self):
        task = Task('(B) something +project1 @context1 pri:C')
        self.assertEqual(task.contexts, ['context1'])
        self.assertEqual(task.projects, ['project1'])
        self.assertFalse(task.is_complete)
        self.assertEqual(task.priority, 'B')
        self.assertEqual(len(task.keywords), 1)

        task.setCompleted()
        self.assertTrue(task.is_complete)
        self.assertEqual(task.completion_date, date.today())
        self.assertTrue(task.text.startswith("x "))

        task.setPending()
        self.assertFalse(task.is_complete)
        self.assertFalse(task.completion_date)

    def test_future(self):
        task = Task('(D) do something +project1 t:2030-10-06')
        self.assertEqual(task.contexts, [])
        self.assertEqual(task.projects, ['project1'])
        self.assertFalse(task.is_complete)
        self.assertTrue(task.priority)
        self.assertTrue(task.is_future)

    def test_custom_keywords(self):
        task = Task('(B) do something +project1 mykey:myval titi:toto @context1 rest + of line pri:C')
        self.assertEqual(task.contexts, ['context1'])
        self.assertEqual(task.projects, ['project1'])
        self.assertFalse(task.is_complete)
        self.assertEqual(task.priority, 'B')
        self.assertEqual(len(task.keywords), 3)
        self.assertIn("titi", task.keywords)
        self.assertEqual(task.keywords['pri'], "C")

    def test_creation_date(self):
        task = Task('(B) 2016-03-15 do something +project1 mykey:myval titi:toto @context1 rest + of line pri:C')
        self.assertTrue(task.creation_date)
        self.assertEqual(task.creation_date, date(2016, 3, 15))
        task = Task('2016-03-15 do something')
        self.assertTrue(task.creation_date)
        self.assertEqual(task.creation_date, date(2016, 3, 15))
        task = Task('do something')
        self.assertFalse(task.creation_date)
        task = Task('(A) do something')
        self.assertFalse(task.creation_date)
        task = Task('x 2017-08-12 2016-03-15 do something')
        self.assertTrue(task.creation_date)
        self.assertEqual(task.creation_date, date(2016, 3, 15))
        task = Task('x 2017-08-12 do something')
        self.assertFalse(task.creation_date)

    # Recurring tasks
    # Negative tests
    def test_recurring_task_wrong_interval(self):
        task = Task('(B) do something rec:2g')
        self.assertIsNone(task.recursion)

    def test_recurring_task_wrong_increment(self):
        task = Task('(B) do something rec:0d')
        self.assertIsNone(task.recursion)

    def test_recurring_task_wrong_keyword(self):
        task = Task('(B) do something rect:5d')
        self.assertIsNone(task.recursion)

    # Positive tests
    def test_recurring_task_input_days(self):
        task = Task('(C) do something due:2016-09-05 rec:5d')
        self.assertIsNotNone(task.recursion)
        self.assertTrue(task.recursion.mode == RecursiveMode.completitionDate)
        self.assertTrue(task.recursion.increment == str(5))
        self.assertTrue(task.recursion.interval == 'd')

    def test_recurring_task_input_weeks(self):
        task = Task('(C) do something due:2016-09-05 rec:+7w')
        self.assertIsNotNone(task.recursion)
        self.assertTrue(task.recursion.mode == RecursiveMode.originalDueDate)
        self.assertTrue(task.recursion.increment == str(7))
        self.assertTrue(task.recursion.interval == 'w')

    def test_recurring_task_input_months(self):
        task = Task('(C) do something due:2016-09-05 rec:3m')
        self.assertIsNotNone(task.recursion)
        self.assertTrue(task.recursion.mode == RecursiveMode.completitionDate)
        self.assertTrue(task.recursion.increment == str(3))
        self.assertTrue(task.recursion.interval == 'm')

    def test_recurring_task_input_years(self):
        task = Task('(C) do something due:2016-09-05 rec:+1y')
        self.assertIsNotNone(task.recursion)
        self.assertTrue(task.recursion.mode == RecursiveMode.originalDueDate)
        self.assertTrue(task.recursion.increment == str(1))
        self.assertTrue(task.recursion.interval == 'y')



from qtodotxt2.lib import task_htmlizer, tasklib
import unittest


class TestHtmlizer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.htmlizer = task_htmlizer.TaskHtmlizer()

    def test_01(self):
        # Simple task should return simple html
        task = tasklib.Task('this is my task')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task')

    def test_02(self):
        # Test task with a single context at the end
        task = tasklib.Task('this is my task @context')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <font style="color:green">@context</font>')

    def test_03(self):
        # Test task with a single context at the center
        task = tasklib.Task('this is my task @context and some more words')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task '
                         '<font style="color:green">@context</font> and some more words')

    def test_04(self):
        # Test task with a single project at the end
        task = tasklib.Task('this is my task +project')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task '
                         '<font style="color:#64AAD0">+project</font>')

    def test_05(self):
        # Test task with a single project at the center
        task = tasklib.Task('this is my task +project and some more words')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task '
                         '<font style="color:#64AAD0">+project</font> and '
                         'some more words')

    def test_06(self):
        # Test task with a single context and a single project
        task = tasklib.Task('this is my task @context and +project and some more words')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task '
                         '<font style="color:green">@context</font> and '
                         '<font style="color:#64AAD0">+project</font> and some more words')

    def test_07(self):
        # Test task with a two contexts and a three projects
        task = tasklib.Task('this is my task @context1 and @context2 and '
                            '+project1 +project2 and +project3 some more '
                            'words')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task '
                         '<font style="color:green">@context1</font> and '
                         '<font style="color:green">@context2</font> and '
                         '<font style="color:#64AAD0">+project1</font> '
                         '<font style="color:#64AAD0">+project2</font> and '
                         '<font style="color:#64AAD0">+project3</font>'
                         ' some more words')

    def test_08(self):
        # Test task with priority A
        task = tasklib.Task('(A) this is my task')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<font style="color:red"><tt>(A)</tt>&nbsp;</font>this is my task')

    def test_09(self):
        # Test task with priority B
        task = tasklib.Task('(B) this is my task')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<font style="color:green"><tt>(B)</tt>&nbsp;</font>this is my task')

    def test_10(self):
        # Test task with priority C
        task = tasklib.Task('(C) this is my task')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<font style="color:navy"><tt>(C)</tt>&nbsp;</font>this is my task')

    def test_11(self):
        # Test task with priority D
        task = tasklib.Task('(D) this is my task')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>(D)</tt>&nbsp;this is my task')

    def test_12(self):
        # Test task with a valid due date
        task = tasklib.Task('this is my task due:2014-04-01')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">due:2014-04-01</font></b>')

    def test_13(self):
        # Test task with a valid due date and time
        task = tasklib.Task('this is my task due:2014-04-01T12:34')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">due:2014-04-01 12:34</font></b>')

    def test_14(self):
        # Test task with an invalid due date
        task = tasklib.Task('this is my task due:abc')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** due:abc Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_15(self):
        # Test task with an invalid due date
        task = tasklib.Task('this is my task due:2014-04')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** due:2014-04 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_16(self):
        # Test task with an invalid due month
        task = tasklib.Task('this is my task due:2014-13-01')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** due:2014-13-01 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_17(self):
        # Test task with an invalid due day
        task = tasklib.Task('this is my task due:2014-04-31')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** due:2014-04-31 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_18(self):
        # Test task with space in due time instead of T. This is valid, but gives an unexpected result
        task = tasklib.Task('this is my task due:2014-04-01 12:34')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">due:2014-04-01</font></b> 12:34')

    def test_19(self):
        # Test task with a valid due time corner case
        task = tasklib.Task('this is my task due:2014-04-01T00:00')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">due:2014-04-01</font></b>')

    def test_20(self):
        # Test task with a valid due time corner case
        task = tasklib.Task('this is my task due:2014-04-01T00:01')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">due:2014-04-01 00:01</font></b>')

    def test_21(self):
        # Test task with a valid due time corner case
        task = tasklib.Task('this is my task due:2014-04-01T23:59')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">due:2014-04-01 23:59</font></b>')

    def test_22(self):
        # Test task with an invalid due time corner case
        task = tasklib.Task('this is my task due:2014-04-01T24:00')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** due:2014-04-01T24:00 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_23(self):
        # Test task with a valid threshold date
        task = tasklib.Task('this is my task t:2014-04-01')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task '
                         '<font style="color:orange">t:2014-04-01</font>')

    def test_24(self):
        # Test task with a valid threshold date and time
        task = tasklib.Task('this is my task t:2014-04-01T12:34')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task '
                         '<font style="color:orange">t:2014-04-01 12:34</font>')

    def test_25(self):
        # Test task with an invalid threshold date
        task = tasklib.Task('this is my task t:abc')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** t:abc Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_26(self):
        # Test task with an invalid threshold date
        task = tasklib.Task('this is my task t:2014-04')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** t:2014-04 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_27(self):
        # Test task with an invalid threshold month
        task = tasklib.Task('this is my task t:2014-13-01')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** t:2014-13-01 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_28(self):
        # Test task with an invalid threshold day
        task = tasklib.Task('this is my task t:2014-04-31')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** t:2014-04-31 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_29(self):
        # Test task with an invalid threshold hour
        task = tasklib.Task('this is my task t:2014-04-31T99:34')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** t:2014-04-31T99:34 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_30(self):
        # Test task with an invalid threshold minute
        task = tasklib.Task('this is my task t:2014-04-31T12:99')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;this is my task <b>'
                         '<font style="color:red">*** t:2014-04-31T12:99 Invalid date format, '
                         'expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>')

    def test_31(self):
        # Test task with an URL
        task = tasklib.Task('Download https://github.com/mNantern/QTodoTxt/archive/master.zip and extract')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;Download '
                         '<a style="color:none;" href="https://github.com/mNantern/QTodoTxt/archive/master.zip">'
                         'https://github.com/...</a> and extract')

    def test_32(self):
        # Test task with solely an URL
        task = tasklib.Task('https://github.com/mNantern/QTodoTxt/archive/master.zip')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;'
                         '<a style="color:none;" href="https://github.com/mNantern/QTodoTxt/archive/master.zip">'
                         'https://github.com/...</a>')

    def test_33(self):
        # Test task with an URL in context
        task = tasklib.Task('Test @https://github.com/mNantern/QTodoTxt/')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;Test <font style="color:green">'
                         '@<a style="color:green;" href="https://github.com/mNantern/QTodoTxt/">'
                         'https://github.com/...</a></font>')

    def test_34(self):
        # Test task with an URL in project
        task = tasklib.Task('Test +https://github.com/mNantern/QTodoTxt/')
        self.assertEqual(self.htmlizer.task2html(task),
                         '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;Test <font style="color:#64AAD0">'
                         '+<a style="color:#64AAD0;" href="https://github.com/mNantern/QTodoTxt/">'
                         'https://github.com/...</a></font>')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import logging
import sys
import os

from PyQt5 import QtCore
from PyQt5 import QtWidgets
from PyQt5.QtQml import QQmlApplicationEngine

import qtodotxt2.qTodoTxt_style_rc

from qtodotxt2.main_controller import MainController
from qtodotxt2.lib.file import FileObserver
from qtodotxt2.lib.tendo_singleton import SingleInstance


def _parseArgs():
    if len(sys.argv) > 1 and sys.argv[1].startswith('-psn'):
        del sys.argv[1]
    parser = argparse.ArgumentParser(description='QTodoTxt')
    parser.add_argument('file', type=str, nargs='?', metavar='TEXTFILE', help='open the specified file')
    parser.add_argument(
        '-l',
        '--loglevel',
        type=str,
        nargs=1,
        metavar='LOGLEVEL',
        default=['WARN'],
        choices=['DEBUG', 'INFO', 'WARNING', 'WARN', 'ERROR', 'CRITICAL'],
        help='set one of these logging levels: DEBUG, INFO, WARNING, ERROR, CRITICAL')
    return parser.parse_args()


def _setupLogging(loglevel):
    numeric_level = getattr(logging, loglevel[0].upper(), None)
    if isinstance(numeric_level, int):
        logging.basicConfig(
            format='{asctime}.{msecs:.0f} [{name}] {levelname}: {message}',
            level=numeric_level,
            style='{',
            datefmt='%H:%M:%S')


def setupAnotherInstanceEvent(controller):
    # Connecting to a processor reading TMP file
    needSingleton = QtCore.QSettings().value("Preferences/singleton", False, type=bool)
    if needSingleton:
        dirname = os.path.dirname(sys.argv[0])
        fileObserver = FileObserver()
        fileObserver.addPath(dirname)
        #FIXME maybe do something in qml
        #fileObserver.dirChangetSig.connect(controller.anotherInstanceEvent)


def setupSingleton(args):
    needSingleton = QtCore.QSettings().value("Preferences/singleton", False, type=bool)
    if int(needSingleton):
        me = SingleInstance()
        dirname = os.path.dirname(sys.argv[0])
        tempFileName = dirname + "/qtodo.tmp"
        if me.initialized is True:
            if os.path.isfile(tempFileName):
                os.remove(tempFileName)
        else:
            f = open(tempFileName, 'w')
            f.flush()
            f.close()
            sys.exit(-1)


def run():
    # First set some application settings for QSettings
    QtCore.QCoreApplication.setOrganizationName("QTodoTxt")
    QtCore.QCoreApplication.setApplicationName("QTodoTxt2")
    # Now set up our application and start
    app = QtWidgets.QApplication(sys.argv)
    # it is said, that this is lighter:
    # (without qwidgets, as we probably don't need them anymore, when transition to qml is done)
    # app = QtGui.QGuiApplication(sys.argv)

    name = QtCore.QLocale.system().name()
    translator = QtCore.QTranslator()
    if translator.load(str(name) + ".qm", "..//i18n"):
        app.installTranslator(translator)

    args = _parseArgs()

    setupSingleton(args)

    _setupLogging(args.loglevel)

    engine = QQmlApplicationEngine(parent=app)
    controller = MainController(args)
    engine.rootContext().setContextProperty("mainController", controller)
    path = os.path.dirname(__file__)
    engine.addImportPath(path + '/qml/')
    engine.load(path + '/qml/QTodoTxt.qml')

    setupAnotherInstanceEvent(controller)

    controller.start()
    app.exec_()
    sys.exit()


if __name__ == "__main__":
    run()



from PyQt5 import QtCore
from PyQt5 import QtGui
from qtodotxt2.lib.filters import ContextFilter, CompleteTasksFilter, DueFilter, DueOverdueFilter, DueThisMonthFilter, \
    DueThisWeekFilter, DueTodayFilter, DueTomorrowFilter, HasContextsFilter, HasDueDateFilter, HasProjectsFilter, \
    ProjectFilter, UncategorizedTasksFilter, AllTasksFilter, PriorityFilter, HasPriorityFilter
from qtodotxt2.lib.filters import SimpleTextFilter, FutureFilter, IncompleteTasksFilter

TotalCountRole = QtCore.Qt.UserRole + 1
CompletedCountRole = QtCore.Qt.UserRole + 2


class FilterItem(QtGui.QStandardItem):
    def __init__(self, parent, strings, flt=None, icon=None, order=None):
        QtGui.QStandardItem.__init__(self, strings)
        self.setSelectable(True)
        self.setEnabled(True)
        self.setData(flt, QtCore.Qt.UserRole)
        self.filter = flt
        parent.appendRow([self])
        #if order:
        #self.setText(1, str(order))
        self.iconSource = icon

    def setCounts(self, total, completed):
        self.setTotalCount(total)
        self.setCompletedCount(completed)

    def setTotalCount(self, total):
        self.setData(total, TotalCountRole)

    def setCompletedCount(self, val):
        self.setData(val, CompletedCountRole)


class FiltersModel(QtGui.QStandardItemModel):
    def __init__(self, parent):
        QtGui.QStandardItemModel.__init__(self, parent)
        self.style = ":/white_icons"
        if str(QtCore.QSettings().value("color_schem", "")).find("dark") >= 0:
            self.style = ":/dark_icons"
        self._filterItemByFilterType = dict()
        self._filterIconByFilterType = dict()
        self._treeItemByFilterType = dict()
        self._addDefaultTreeItems()
        self._initFilterTypeMappings()

    def roleNames(self):
        roles = QtGui.QStandardItemModel.roleNames(self)
        roles[TotalCountRole] = b"totalCount"
        roles[CompletedCountRole] = b"completedCount"
        return roles

    def addFilter(self, flt, counts):
        parent = self._filterItemByFilterType[type(flt)]
        item = FilterItem(parent, flt.text, flt)
        item.setCounts(*counts)

    def clear(self):
        QtGui.QStandardItemModel.clear(self)
        self._addDefaultTreeItems()
        self._initFilterTypeMappings()

    def _addDefaultTreeItems(self):
        self._allTasksItem = FilterItem(self, 'All', AllTasksFilter(), 'qtodotxt-filter-all')
        self._uncategorizedTasksItem = FilterItem(self, 'Uncategorized',
                                                  UncategorizedTasksFilter(), 'qtodotxt-filter-uncategorized')
        self._dueItem = FilterItem(self, 'Due', HasDueDateFilter(), 'qtodotxt-filter-due')
        self._contextsItem = FilterItem(self, 'Contexts', HasContextsFilter(), 'qtodotxt-filter-contexts')
        self._projectsItem = FilterItem(self, 'Projects', HasProjectsFilter(), 'qtodotxt-filter-projects')
        self._priorityItem = FilterItem(self, 'Priorities', HasPriorityFilter(), 'qtodotxt-filter-complete')
        self._completeTasksItem = FilterItem(self, 'Complete', CompleteTasksFilter(), 'qtodotxt-filter-complete')

    def _initFilterTypeMappings(self):
        self._filterItemByFilterType[ContextFilter] = self._contextsItem
        self._filterItemByFilterType[ProjectFilter] = self._projectsItem
        self._filterItemByFilterType[DueFilter] = self._dueItem
        self._filterItemByFilterType[PriorityFilter] = self._priorityItem

        self._filterIconByFilterType[ContextFilter] = 'qtodotxt-filter-contexts'
        self._filterIconByFilterType[ProjectFilter] = 'qtodotxt-filter-projects'

        self._filterIconByFilterType[DueTodayFilter] = 'qtodotxt-filter-due-today'
        self._filterIconByFilterType[DueTomorrowFilter] = 'qtodotxt-filter-due-tomorrow'
        self._filterIconByFilterType[DueThisWeekFilter] = 'qtodotxt-filter-due-week'
        self._filterIconByFilterType[DueThisMonthFilter] = 'qtodotxt-filter-due-month'
        self._filterIconByFilterType[DueOverdueFilter] = 'qtodotxt-filter-due-overdue'
        self._filterIconByFilterType[PriorityFilter] = 'qtodotxt-filter-complete'

        self._treeItemByFilterType[AllTasksFilter] = self._allTasksItem
        self._treeItemByFilterType[UncategorizedTasksFilter] = self._uncategorizedTasksItem
        self._treeItemByFilterType[CompleteTasksFilter] = self._completeTasksItem
        self._treeItemByFilterType[HasProjectsFilter] = self._projectsItem
        self._treeItemByFilterType[HasDueDateFilter] = self._dueItem
        self._treeItemByFilterType[HasContextsFilter] = self._contextsItem
        self._treeItemByFilterType[HasPriorityFilter] = self._priorityItem

    # Predefined sorting for due ranges
    def addDueRangeFilter(self, flt, counts, sortKey=0):
        parentItem = self._dueItem
        icon = self._filterIconByFilterType[type(flt)]
        item = FilterItem(parentItem, flt.text, flt=flt, icon=icon, order=sortKey)
        item.setCounts(*counts)

        #parentItem.setExpanded(True)
        #parentItem.sortChildren(1, QtCore.Qt.AscendingOrder)

    @QtCore.pyqtSlot(result='QVariantList')
    def getRootChildren(self):
        indexes = []
        parent = self.invisibleRootItem()
        for i in range(0, parent.rowCount()):
            child = parent.child(i)
            indexes.append(child.index())
        return indexes

    @QtCore.pyqtSlot('QModelIndex', result='QString')
    def iconFromIndex(self, index):
        source = ""
        item = super(FiltersModel, self).itemFromIndex(index)
        if item is not None:
            source = item.iconSource
        return source

    @QtCore.pyqtSlot('int', result='QString')
    def iconFromRow(self, row):
        path = ""
        if row > 0 and row < self.rowCount():
            path = self.item(row, 0).iconSource
        return path

    def updateCounters(self, counters):
        self._completeTasksItem.setTotalCount(counters['All'][1])
        self._allTasksItem.setCounts(*counters['All'])
        self._dueItem.setCounts(*counters['Due'])
        self._contextsItem.setCounts(*counters['Contexts'])
        self._projectsItem.setCounts(*counters['Projects'])
        self._priorityItem.setCounts(*counters['Priority'])
        self._uncategorizedTasksItem.setCounts(*counters['Uncategorized'])


class FiltersController(QtCore.QObject):

    filterSelectionChanged = QtCore.pyqtSignal(list)

    def __init__(self):
        QtCore.QObject.__init__(self)
        self._settings = QtCore.QSettings()
        self.model = FiltersModel(self)
        self.showCompleted = self._settings.value("show_completed", False)
        self.showFuture = self._settings.value("show_completed", True)
        self.searchText = ""
        # self.currentFilters = self._settings.value("current_filters", ["All"])  # move to QML
        self.currentFilters = []

    def setFiltersByIndexes(self, idxs):
        filters = [self.model.itemFromIndex(idx).filter for idx in idxs]
        self.setFilters(filters)

    def setFilters(self, filters):
        self.currentFilters = filters

    def filter(self, tasks):
        # First we filter with filters tree
        tasks = filterTasks(self.currentFilters, tasks)
        # Then with our search text
        if self.searchText:
            tasks = filterTasks([SimpleTextFilter(self.searchText)], tasks)
        # with future filter if needed
        if not self.showFuture:
            tasks = filterTasks([FutureFilter()], tasks)
        # with complete filter if needed
        if not self.showCompleted and CompleteTasksFilter() not in self.currentFilters:
            tasks = filterTasks([IncompleteTasksFilter()], tasks)
        return tasks

    def updateFiltersModel(self, mfile):
        self.model.clear()
        self._addAllContexts(mfile)
        self._addAllProjects(mfile)
        self._addAllDueRanges(mfile)
        self._addAllPriorities(mfile)
        self._updateCounter(mfile)

    def _updateCounter(self, mfile):
        counters = mfile.getTasksCounters()
        self.model.updateCounters(counters)

    def _addAllContexts(self, mfile):
        contexts = mfile.getAllContexts()
        for context, number in contexts.items():
            mfilter = ContextFilter(context)
            self.model.addFilter(mfilter, number)

    def _addAllProjects(self, mfile):
        projects = mfile.getAllProjects()
        for project, counts in projects.items():
            mfilter = ProjectFilter(project)
            self.model.addFilter(mfilter, counts)

    def _addAllPriorities(self, mfile):
        priorities = mfile.getAllPriorities()
        for priority, number in priorities.items():
            mfilter = PriorityFilter(priority)
            self.model.addFilter(mfilter, number)

    def _addAllDueRanges(self, mfile):
        dueRanges = mfile.getAllDueRanges()

        for flt, counts in dueRanges.items():
            self.model.addDueRangeFilter(flt, counts)


def filterTasks(filters, tasks):
    if not filters:
        return tasks[:]

    filteredTasks = []
    for task in tasks:
        for myfilter in filters:
            if myfilter.isMatch(task):
                filteredTasks.append(task)
                break
    return filteredTasks





import re
from datetime import date, datetime, timedelta


class BaseFilter(object):
    """
    The abstract base class for different kind of task-list filters.

    """

    def __init__(self, text):
        """
        Initialize a new BaseFilter objects.

        The required text argument (str) becomes the "text" instance attribute
        of the object.

        """
        self.text = text

    def __hash__(self):
        return hash(self.text)

    def __str__(self):
        return "Filter:{}".format(self.__class__.__name__)

    __repr__ = __str__

    def isMatch(self, task):
        """
        Determine whether the supplied task (arg 'task') satisfies the filter.

        In this base class, the test always returns True.

        """
        return True

    def __eq__(self, other):
        """
        Evaluates objects as equal if their type and self.text attr are the same.

        """
        if not other:
            return False
        if type(other) == type(self):
            return self.text == other.text
        return False


class AllTasksFilter(BaseFilter):
    """
    Task list filter that returns all tasks.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'All')


class IncompleteTasksFilter(BaseFilter):
    """
    Task list filter that removes any completed tasks.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'Incomplete')

    def isMatch(self, task):
        return not task.is_complete


class UncategorizedTasksFilter(BaseFilter):
    """
    Task list filter permitting incomplete tasks without project or context.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'Uncategorized')

    def isMatch(self, task):
        return (not task.contexts) and (not task.projects)


class CompleteTasksFilter(BaseFilter):
    """
    Task list filter that removes any uncompleted tasks from the list.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'Complete')

    def isMatch(self, task):
        return task.is_complete


class ContextFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks with the selected context.

    """

    def __init__(self, context):
        BaseFilter.__init__(self, context)

    def isMatch(self, task):
        return self.text in task.contexts

    def __str__(self):
        return "ContextFilter(%s)" % self.text


class ProjectFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks with the selected project.

    """

    def __init__(self, project):
        BaseFilter.__init__(self, project)

    def isMatch(self, task):
        return self.text in task.projects

    def __str__(self):
        return "ProjectFilter(%s)" % self.text


class DueFilter(BaseFilter):
    """
    Due list filter for ranges

    """

    def __init__(self, dueRange):
        BaseFilter.__init__(self, dueRange)

    def isMatch(self, task):
        return self.text in task.dueRanges

    def __str__(self):
        return "DueFilter(%s)" % self.text


class DueTodayFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks that are due today.

    """

    def __init__(self):
        BaseFilter.__init__(self, "Today")

    def isMatch(self, task):
        if (not task.due):
            return False
        else:
            self.due_date = task.due
            today = datetime.combine(date.today(), datetime.min.time())
            return self.due_date == today

    def __str__(self):
        return "DueTodayFilter(%s)" % self.text


class DueTomorrowFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks that are due tomorrow.

    """

    def __init__(self):
        BaseFilter.__init__(self, "Tomorrow")

    def isMatch(self, task):
        if not task.due:
            return False
        else:
            due_date = task.due
            today = datetime.combine(date.today(), datetime.min.time())
            return today < due_date <= today + timedelta(days=1)

    def __str__(self):
        return "DueTomorrowFilter(%s)" % self.text


class DueThisWeekFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks that are due this week.

    """

    def __init__(self):
        BaseFilter.__init__(self, "This week")

    def isMatch(self, task):
        if not task.due:
            return False
        else:
            due_date = task.due
            today = datetime.combine(date.today(), datetime.min.time())
            return today <= due_date <= today + timedelta((6 - today.weekday()) % 7)

    def __str__(self):
        return "DueThisWeekFilter(%s)" % self.text


class DueThisMonthFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks that are due this month.

    """

    def __init__(self):
        BaseFilter.__init__(self, "This month")

    def isMatch(self, task):
        if not task.due:
            return False
        else:
            due_date = task.due
            today = datetime.combine(date.today(), datetime.min.time())
            if today.month == 12:
                last_day_of_month = today.replace(day=31)
            else:
                last_day_of_month = today.replace(month=today.month + 1, day=1) - timedelta(days=1)
            return today <= due_date <= last_day_of_month

    def __str__(self):
        return "DueThisMonthFilter(%s)" % self.text


class DueOverdueFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks that are overdue.

    """

    def __init__(self):
        BaseFilter.__init__(self, "Overdue")

    def isMatch(self, task):
        if not task.due:
            return False
        else:
            due_date = task.due
            today = datetime.combine(date.today(), datetime.min.time())
            return due_date < today

    def __str__(self):
        return "DueOverdueFilter(%s)" % self.text


class HasProjectsFilter(BaseFilter):
    """
    Task list filter allowing only incomplete tasks with the selected project.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'Projects')

    def isMatch(self, task):
        return task.projects

    def __str__(self):
        return "HasProjectsFilter" % self.text


class HasContextsFilter(BaseFilter):
    """
    Task list filter allowing only tasks tagged with some project.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'Contexts')

    def isMatch(self, task):
        return task.contexts

    def __str__(self):
        return "HasContextsFilter" % self.text


class HasPriorityFilter(BaseFilter):
    """
    Task list filter allowing only tasks with a priority set

    """

    def __init__(self):
        BaseFilter.__init__(self, 'Priorities')

    def isMatch(self, task):
        return task.priority

    def __str__(self):
        return "HasPriorityFilter" % self.text


class HasDueDateFilter(BaseFilter):
    """
    Task list filter allowing only complete tasks with due date in due ranges.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'DueRange')

    def isMatch(self, task):
        return task.due

    def __str__(self):
        return "HasDueDateFilter" % self.text


class HasDueRangesFilter(BaseFilter):
    """
    Task list filter allowing only complete tasks with due date in due ranges.

    """

    def __init__(self):
        BaseFilter.__init__(self, 'DueRange')

    def isMatch(self, task):
        return task.dueRanges

    def __str__(self):
        return "HasDueRangesFilter" % self.text


class SimpleTextFilter(BaseFilter):
    r"""
    Task list filter allowing only tasks whose string matches a filter string.

    This filter allows for basic and/or/not conditions in the filter string.
    For the syntax see SimpleTextFilter.isMatch.

    User documentation:

    This filter can handle basic and/or/not conditions. The syntax is as
    follows:

    :AND   :   ',' or whitespace (' ')
    :OR    :   '|'
    :NOT   :   prefixed '~' or '!'

    These operators follow the following order of precedence: OR, AND, NOT.
    So, for example:

    :'work job1 | @home':                Either  (matches 'work'
                                                 AND 'job1')
                                        OR      (matches '@home')

    :'norweigan blue ~dead | !parrot':  Either  (matches 'norweigan'
                                                 AND 'blue'
                                                 AND does NOT match 'dead')
                                        OR      (does NOT match 'parrot')

    Terms match the beginning of words, so:

    - 'cleese' Does NOT match 'johncleese'

    Since the python re module is used, most of the escaped regex
    characters will also work when attached to one of the (comma- or space-
    delimited) strings. E.g.:

    - john\b will match 'john' but not 'johncleese'
    - 2014-\d\d-07 will match '2014-03-07' but not '2014-ja-07'

    - '(B)' will match '(B) nail its feet to the perch'.
    """

    def __init__(self, text):
        BaseFilter.__init__(self, text)
        self.re = SimpleTextFilter.compile(text)

    # Terms are split by "," " " or "|".  Retain only a "|" for the re.
    _splitter = re.compile(r'\s*(\|)\s*|[,\s]+()', re.U)
    # Escape anything that's not alphanumeric or a backslash
    _escaper = re.compile(r'(\\\Z|[^\w\\])', re.U)
    # Characters that negate a term
    _negates = ('!', '~')

    @staticmethod
    def _term2re(term):
        # Don't translate separators
        if term is None or term == '|' or term == '':
            return term or ''

        # Check for negated  term
        negate = term.startswith(SimpleTextFilter._negates)
        if negate:
            term = term[1:]

        # If the term is a word (starts with alpha), then only match beginnings of words
        beginning = r'\b' if re.match(r'\b', term) else r'\B'

        # Ignore special meaning of characters like "*", "+", "(", "[", etc.
        term = SimpleTextFilter._escaper.sub(r'\\\1', term)

        # Return a pattern that will match the beginning of a word or not,
        # anywhere in the string, without consuming any of the string.
        return r'^(?' + ('!' if negate else '=') + r'.*' + beginning + term + r')'

    @staticmethod
    def compile(searchString):
        r"""
        Return the user's searchString compiled to a regular expression.

        Example terms: @call +work (A) carrots
        Term may be prefixed with ! or ~ for negation.
        Terms may be combined with "," or " " (AND) or with "|" (OR).
        Terms only match the beginning of a word in the task.
        Terms are case-insensitive.
        Expressions may NOT be nested with parentheses.
        Only \-character special regular expression sets are allowed, everything else is escaped.
        """
        if not searchString:
            return None

        terms = SimpleTextFilter._splitter.split(searchString)
        terms = [SimpleTextFilter._term2re(term) for term in terms]

        return re.compile("".join(terms), re.I | re.U)

    def isMatch(self, task):
        """ Return a boolean based on whether the supplied task satisfies self.text. """
        return self.re.match(task.text) if self.re else True

    def __str__(self):
        return "SimpleTextFilter({})".format(self.text)


class FutureFilter(BaseFilter):
    def __init__(self):
        BaseFilter.__init__(self, 'Future')

    def isMatch(self, task):
        return not task.is_future

    def __str__(self):
        return "FutureFilter " % self.text


class PriorityFilter(BaseFilter):
    """
    Task list filter allowing only tasks with a certain priority

    """

    def __init__(self, priority):
        BaseFilter.__init__(self, priority)

    def isMatch(self, task):
        return self.text in task.priority

    def __str__(self):
        return "PriorityFilter " % self.text



from datetime import datetime, date, time, MAXYEAR
import re
from enum import Enum

from PyQt5 import QtCore

from qtodotxt2.lib.task_htmlizer import TaskHtmlizer


class RecursiveMode(Enum):
    completitionDate = 0  # Original due date mode: Task recurs from original due date
    originalDueDate = 1  # Completion date mode: Task recurs from completion date


class Recursion:
    mode = None
    increment = None
    interval = None

    def __init__(self, arg_mode, arg_increment, arg_interval):
        self.mode = arg_mode
        self.increment = arg_increment
        self.interval = arg_interval


class TaskSorter(object):
    
    @staticmethod
    def projects(tasks):
        def tmp(task):
            prj = task.projects if task.projects else ["zz"]
            return prj, task
        return sorted(tasks, key=tmp)

    @staticmethod
    def contexts(tasks):
        def tmp(task):
            ctx = task.contexts if task.contexts else ["zz"]
            return ctx, task
        return sorted(tasks, key=tmp)

    @staticmethod
    def due(tasks):
        def tmp(task):
            if task.due:
                return task.due, task
            else:
                return datetime(MAXYEAR, 1, 1), task
        return sorted(tasks, key=tmp, reverse=False)

    @staticmethod
    def default(tasks):
        return sorted(tasks, reverse=False)


class Task(QtCore.QObject):
    """
    Represent a task as defined in todo.txt format
    Take a line in todo.txt format as argument
    Arguments are read-only, reparse string to modify them or
    use one the modification methods such as setCompleted()
    """

    modified = QtCore.pyqtSignal(object)

    def __init__(self, text):
        QtCore.QObject.__init__(self)
        self._settings = QtCore.QSettings()
        self._highest_priority = 'A'
        # all other class attributes are defined in _reset method
        # which is called in _parse
        self._parse(text)

    def addCreationCate(self):
        self._removeCreationDate()
        self._addCreationDate()

    def __str__(self):
        return self._text

    def __repr__(self):
        return "Task({})".format(self._text)

    def _removeCreationDate(self):
        match = re.match(r'^(\([A-Z]\)\s)?[0-9]{4}\-[0-9]{2}\-[0-9]{2}\s(.*)', self._text)
        if match:
            if match.group(1):
                self._text = match.group(1) + match.group(2)
            else:
                self._text = match.group(2)

    def addCreationDate(self):
        date_string = date.today().strftime('%Y-%m-%d')
        if re.match(r'^\([A-Z]\)', self._text):
            self._text = '%s %s %s' % (self._text[:3], date_string, self._text[4:])
        else:
            self._text = '%s %s' % (date_string, self._text)

    def _reset(self):
        self.contexts = []
        self.projects = []
        self._priority = ""
        self.is_complete = False
        self.completion_date = None
        self.creation_date = None
        self.is_future = False
        self.threshold_error = ""
        self._text = ''
        self.description = ''
        self.due = None
        self.due_error = ""
        self.threshold = None
        self.keywords = {}
        self.recursion = None

    def _parse(self, line):
        """
        parse a task formated as string in todo.txt format
        """
        self._reset()
        words = line.split(' ')
        if words[0] == "x":
            self.is_complete = True
            words = words[1:]
            # parse next word as a completion date
            # required by todotxt but often not here
            self.completion_date = _parseDate(words[0])
            if self.completion_date:
                words = words[1:]
        elif re.search(r'^\([A-Z]\)$', words[0]):
            self._priority = words[0][1:-1]
            words = words[1:]

        dato = _parseDate(words[0])
        if dato:
            self.creation_date = dato
            words = words[1:]

        self.description = " ".join(words)
        for word in words:
            self._parseWord(word)
        self._text = line

    @QtCore.pyqtProperty('QString', notify=modified)
    def text(self):
        return self._text

    @text.setter
    def text(self, txt):
        self._parse(txt)
        self.modified.emit(self)

    @QtCore.pyqtProperty('QString', notify=modified)
    def html(self):
        return self.toHtml()

    @QtCore.pyqtProperty('QString', notify=modified)
    def priority(self):
        return self._priority

    @QtCore.pyqtProperty('QString', notify=modified)
    def priorityHtml(self):
        htmlizer = TaskHtmlizer()
        return htmlizer._htmlizePriority(self.priority)

    def _parseWord(self, word):
        if len(word) > 1:
            if word.startswith('@'):
                self.contexts.append(word[1:])
            elif word.startswith('+'):
                self.projects.append(word[1:])
            elif ":" in word:
                self._parseKeyword(word)

    def _parseKeyword(self, word):
        key, val = word.split(":", 1)
        self.keywords[key] = val
        if word.startswith('due:'):
            self.due = _parseDateTime(word[4:])
            if not self.due:
                print("Error parsing due date '{}'".format(word))
                self.due_error = word[4:]
        elif word.startswith('t:'):
            self._parseFuture(word)
        elif word.startswith('rec:'):
            self._parseRecurrence(word)

    def _parseFuture(self, word):
        self.threshold = _parseDateTime(word[2:])
        if not self.threshold:
            print("Error parsing threshold '{}'".format(word))
            self.threshold_error = word[2:]
        else:
            if self.threshold > datetime.today():
                self.is_future = True

    def _parseRecurrence(self, word):
        # Original due date mode
        if word[4] == '+':
            # Test if chracters have the right format
            if re.match('^[1-9][bdwmy]', word[5:7]):
                self.recursion = Recursion(RecursiveMode.originalDueDate, word[5], word[6])
            else:
                print("Error parsing recurrence '{}'".format(word))
        # Completion mode
        else:
            # Test if chracters have the right format
            if re.match('^[1-9][bdwmy]', word[4:6]):
                self.recursion = Recursion(RecursiveMode.completitionDate, word[4], word[5])
            else:
                print("Error parsing recurrence '{}'".format(word))

    @property
    def dueString(self):
        return dateString(self.due)

    @property
    def thresholdString(self):
        return dateString(self.threshold)

    @QtCore.pyqtSlot()
    def toggleCompletion(self):
        if self.is_complete:
            self.setPending()
        else:
            self.setCompleted()

    def setCompleted(self):
        """
        Set a task as completed by inserting a x and current date at the begynning of line
        """
        if self.is_complete:
            return
        self.completion_date = date.today()
        date_string = self.completion_date.strftime('%Y-%m-%d')
        self._text = 'x %s %s' % (date_string, self._text)
        self.is_complete = True
        self.modified.emit(self)

    def setPending(self):
        """
        Unset completed flag from task
        """
        if not self.is_complete:
            return
        words = self._text.split(" ")
        d = _parseDate(words[1])
        if d:
            self._text = " ".join(words[2:])
        else:
            self._text = " ".join(words[1:])
        self.is_complete = False
        self.completion_date = None
        self.modified.emit(self)

    def toHtml(self):
        """
        return a task as an html block which is a pretty display of a line in todo.txt format
        """
        htmlizer = TaskHtmlizer()
        return htmlizer.task2html(self)

    def _getLowestPriority(self):
        return self._settings.value("Preferences/lowest_priority", "D")

    @QtCore.pyqtSlot()
    def increasePriority(self):
        lowest_priority = self._getLowestPriority()
        if self.is_complete:
            return
        if not self._priority:
            self._priority = lowest_priority
            self._text = "({}) {}".format(self._priority, self._text)
        elif self._priority != self._highest_priority:
            self._priority = chr(ord(self._priority) - 1)
            self._text = "({}) {}".format(self._priority, self._text[4:])
        self.modified.emit(self)

    @QtCore.pyqtSlot()
    def decreasePriority(self):
        lowest_priority = self._getLowestPriority()
        if self.is_complete:
            return
        if self._priority >= lowest_priority:
            self._priority = ""
            self._text = self._text[4:]
            self._text = self._text.replace("({})".format(self._priority), "", 1)
        elif self._priority:
            oldpriority = self._priority
            self._priority = chr(ord(self._priority) + 1)
            self._text = self._text.replace("({})".format(oldpriority), "({})".format(self._priority), 1)
        self.modified.emit(self)

    def __eq__(self, other):
        return self._text == other.text

    def __lt__(self, other):
        prio1 = self.priority if self.priority else "z"
        prio2 = other.priority if other.priority else "z"
        return (self.is_complete, prio1, self._text) < (other.is_complete, prio2, other.text)


def dateString(date):
    """
    Return a datetime as a nicely formatted string
    """
    if date.time() == time.min:
        return date.strftime('%Y-%m-%d')
    else:
        return date.strftime('%Y-%m-%d %H:%M')


def updateDateInTask(text, newDate):
    # (A) 2016-12-08 Feed Schrodinger's Cat rec:9w due:2016-11-23
    text = re.sub(r'\sdue\:[0-9]{4}\-[0-9]{2}\-[0-9]{2}', ' due:' + str(newDate)[0:10], text)
    return text


def _parseDate(string):
    try:
        return datetime.strptime(string, '%Y-%m-%d').date()
    except ValueError:
        return None


def _parseDateTime(string):
    try:
        return datetime.strptime(string, '%Y-%m-%d')
    except ValueError:
        try:
            return datetime.strptime(string, '%Y-%m-%dT%H:%M')
        except ValueError:
            return None





import logging
import os
import string

from PyQt5 import QtCore

from qtodotxt2.lib import tasklib
from qtodotxt2.lib.file import File
from qtodotxt2.filters_controller import FiltersController

logger = logging.getLogger(__name__)


class MainController(QtCore.QObject):

    error = QtCore.pyqtSignal(str, arguments=["msg"])
    fileExternallyModified = QtCore.pyqtSignal()

    def __init__(self, args):
        super(MainController, self).__init__()
        self._args = args
        self._filteredTasks = []
        self._sortingMode = "default"
        # use object variable for setting only used in this class
        # others are accessed through QSettings
        self._settings = QtCore.QSettings()
        self._file = File()
        self._file.fileExternallyModified.connect(self.fileExternallyModified)
        self._file.fileModified.connect(self._fileModified)
        self.filtersController = FiltersController()
        self._title = "QTodoTxt"
        self._recentFiles = self._settings.value("recent_files", [])
        self._updateCompletionStrings()
        self._forced = None

    def showError(self, msg):
        logger.debug("ERROR: %s", msg)
        self.error.emit(msg)

    completionChanged = QtCore.pyqtSignal()

    @QtCore.pyqtProperty('QStringList', notify=completionChanged)
    def completionStrings(self):
        return self._completionStrings

    def _updateCompletionStrings(self):
        contexts = ['@' + name for name in self._file.getAllContexts()]
        projects = ['+' + name for name in self._file.getAllProjects()]
        lowest_priority = self._settings.value("lowest_priority", "D")
        idx = string.ascii_uppercase.index(lowest_priority) + 1
        priorities = ['(' + val + ')' for val in string.ascii_uppercase[:idx]]
        self._completionStrings = contexts + projects + priorities + ['due:']
        self.completionChanged.emit()

    @QtCore.pyqtSlot('QModelIndexList')
    def filterByIndexes(self, idxs):
        self.filtersController.setFiltersByIndexes(idxs)
        self.applyFilters()

    @QtCore.pyqtSlot('QString', 'int', result='int')
    def newTask(self, text='', after=None):
        task = tasklib.Task(text)
        if bool(self._settings.value("Preferences/add_creation_date", False, type=bool)):
            task.addCreationDate()
        if after is None:
            after = len(self._filteredTasks) - 1
        #self._file.addTask(task)
        self._filteredTasks.insert(after + 1, task)  # force the new task to be visible
        self._file.tasks.append(task)

        self._file.connectTask(task)  #Ensure task will be added
        self.filteredTasksChanged.emit()
        return after + 1

    @QtCore.pyqtSlot('QVariant')
    def deleteTask(self, task):
        if not isinstance(task, tasklib.Task):
            # if task is not a task assume it is an int
            task = self._filteredTasks[task]
        self._file.deleteTask(task)

    @property
    def allTasks(self):
        return self._file.tasks

    @allTasks.setter
    def allTasks(self, tasks):
        self._file.tasks = tasks

    filteredTasksChanged = QtCore.pyqtSignal()

    @QtCore.pyqtProperty('QVariant', notify=filteredTasksChanged)
    def filteredTasks(self):
        return self._filteredTasks

    showFutureChanged = QtCore.pyqtSignal('bool')

    @QtCore.pyqtProperty('bool', notify=showFutureChanged)
    def showFuture(self):
        return self.filtersController.showFuture

    @showFuture.setter
    def showFuture(self, val):
        self.filtersController.showFuture = val
        self.showFutureChanged.emit(val)
        self.applyFilters()

    sortingModeChanged = QtCore.pyqtSignal(str)

    @QtCore.pyqtProperty(str, notify=showFutureChanged)
    def sortingMode(self):
        return self._sortingMode

    @sortingMode.setter
    def sortingMode(self, val):
        self._sortingMode = val
        self.sortingModeChanged.emit(val)
        self.applyFilters()

    searchTextChanged = QtCore.pyqtSignal(str)

    @QtCore.pyqtProperty('QString', notify=searchTextChanged)
    def searchText(self):
        return self.filtersController.searchText

    @searchText.setter
    def searchText(self, txt):
        self.filtersController.searchText = txt
        self.applyFilters()
        self.searchTextChanged.emit(txt)

    showCompletedChanged = QtCore.pyqtSignal('bool')

    @QtCore.pyqtProperty('bool', notify=showCompletedChanged)
    def showCompleted(self):
        return self.filtersController.showCompleted

    @showCompleted.setter
    def showCompleted(self, val):
        self.filtersController.showCompleted = val
        self.showCompletedChanged.emit(val)
        self.applyFilters()

    def auto_save(self):
        if bool(self._settings.value("Preferences/auto_save", True, type=bool)):
            self.save()

    def start(self):
        if self._args.file:
            filename = self._args.file
        else:
            filename = self._settings.value("last_open_file")

        if filename:
            try:
                self.open(filename)
            except OSError as ex:
                self.showError(str(ex))

        self.applyFilters()
        self._updateTitle()

    filtersUpdated = QtCore.pyqtSignal()

    @QtCore.pyqtProperty('QVariant', notify=filtersUpdated)
    def filtersModel(self):
        return self.filtersController.model

    def _updateFilterTree(self):
        self.filtersController.updateFiltersModel(self._file)
        self.filtersUpdated.emit()

    def applyFilters(self, filters=None):
        if filters is not None:
            self.filtersController.setFilters(filters)
        tasks = self.filtersController.filter(self._file.tasks)
        tasks = getattr(tasklib.TaskSorter, self._sortingMode)(tasks)
        self._filteredTasks = tasks
        self.filteredTasksChanged.emit()

    @QtCore.pyqtSlot()
    def archiveCompletedTasks(self):
        done = [task for task in self._file.tasks if task.is_complete]
        for task in done:
            self._file.saveDoneTask(task)
            self._file.deleteTasl(task)
        self.applyFilters()
        self.auto_save()

    modifiedChanged = QtCore.pyqtSignal(bool)

    @QtCore.pyqtProperty('bool', notify=modifiedChanged)
    def modified(self):
        return self._file.modified

    def _fileModified(self, val=True):
        if val:
            self.auto_save()
            self.applyFilters()
            self._updateCompletionStrings()
            self._updateFilterTree()
        self._updateTitle()
        self.modifiedChanged.emit(val)

    @QtCore.pyqtSlot("QUrl")
    @QtCore.pyqtSlot()
    def save(self, path=None):
        if not path:
            path = self._file.filename
        elif isinstance(path, QtCore.QUrl):
            path = path.toLocalFile()
        self._file.filename = path

        logger.debug('MainController, saving file: %s.', path)
        try:
            self._file.save(path)
        except OSError as ex:
            logger.exception("Error saving file %s", path)
            self.showError(ex)
            return
        self._settings.setValue("last_open_file", path)
        self._settings.sync()

    def _updateTitle(self):
        title = 'QTodoTxt - '
        if self._file.filename:
            filename = os.path.basename(self._file.filename)
            title += filename
        else:
            title += 'Untitled'
        if self._file.modified:
            title += ' (*)'
        self._title = title
        self.titleChanged.emit(self._title)

    titleChanged = QtCore.pyqtSignal(str)

    @QtCore.pyqtProperty('QString', notify=titleChanged)
    def title(self):
        return self._title

    @QtCore.pyqtSlot(result='bool')
    def canExit(self):
        self.auto_save()
        return not self._file.modified

    def new(self):
        if self.canExit():
            self._file = File()
            self._loadFileToUI()

    @QtCore.pyqtSlot()
    def reload(self):
        self.open(self._file.filename)

    @QtCore.pyqtSlot('QUrl')
    @QtCore.pyqtSlot('QString')
    def open(self, filename):
        if isinstance(filename, QtCore.QUrl):
            filename = filename.toLocalFile()
        logger.debug('MainController.open called with filename="%s"', filename)
        try:
            self._file.load(filename)
        except Exception as ex:
            self.showError(self.tr("Error opening file: {}.\n Exception:{}").format(filename, ex))
            return
        self._loadFileToUI()
        self._settings.setValue("last_open_file", filename)
        self.updateRecentFile()

    recentFilesChanged = QtCore.pyqtSignal()

    @QtCore.pyqtProperty('QVariant', notify=recentFilesChanged)
    def recentFiles(self):
        return self._recentFiles

    def updateRecentFile(self):
        if self._file.filename in self._recentFiles:
            self._recentFiles.remove(self._file.filename)
        self._recentFiles.insert(0, self._file.filename)
        self._recentFiles = self._recentFiles[:int(self._settings.value("max_recent_files", 6))]
        self._settings.setValue("recent_files", self._recentFiles)
        self.recentFilesChanged.emit()

    def _loadFileToUI(self):
        self.applyFilters()
        self._updateCompletionStrings()
        self._updateFilterTree()



from setuptools import setup, find_packages

import sys


setup(name="qtodotxt2", 
      version="2.0.0a3",
      description="Cross Platform todo.txt GUI",
      author="QTT Development Team",
      author_email="qtodotxt@googlegroups.com",
      url='https://github.com/QTodoTxt/QTodoTxt2',
      #packages=find_packages(where='.', include=["*.py", "*.qrc", "*.qml"], exclude=["tests"]),
      packages=find_packages(),
      #packages=['qtodotxt', 'qtodotxt/lib'],
      package_data={
          'qtodotxt2':['qml/*.qml', 'qml/Theme/*.qml'],
          },
      #include_package_data=True,
      provides=["qtodotxt2"],
      install_requires=["python-dateutil"],
      license="GNU General Public License v3 or later",
      classifiers=["Environment :: X11 Applications :: Qt",
                   "Programming Language :: Python :: 3",
                   "Intended Audience :: End Users/Desktop",
                   "Operating System :: OS Independent",
                   "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)"
                   ],
      entry_points={'gui_scripts': 
                    [
                        'qtodotxt2 = qtodotxt2.app:run'
                    ]
                    }
      )




import logging
import os

from PyQt5 import QtCore

from qtodotxt2.lib.filters import DueTodayFilter, DueTomorrowFilter, DueThisWeekFilter, DueThisMonthFilter, DueOverdueFilter
from qtodotxt2.lib.tasklib import Task

logger = logging.getLogger(__name__)


class File(QtCore.QObject):

    fileExternallyModified = QtCore.pyqtSignal()
    fileModified = QtCore.pyqtSignal(bool)

    def __init__(self):
        QtCore.QObject.__init__(self)
        self.newline = '\n'
        self.tasks = []
        self.filename = ''
        self._fileObserver = FileObserver()
        self._fileObserver.fileChangetSig.connect(self.fileExternallyModified)
        self.modified = False

    def __str__(self):
        return "File(filename:{}, tasks:{})".format(self.filename, self.tasks)

    __repr__ = __str__

    def load(self, filename):
        self._fileObserver.clear()
        with open(filename, 'rt', encoding='utf-8') as fd:
            lines = fd.readlines()
        self.filename = filename
        self._createTasksFromLines(lines)
        self._fileObserver.addPath(self.filename)

    def _createTasksFromLines(self, lines):
        self.tasks = []
        for line in lines:
            task_text = line.strip()
            if task_text:
                task = Task(task_text)
                self.tasks.append(task)
                task.modified.connect(self._taskModified)

    def _taskModified(self, task):
        self.setModified(True)
        #if task not in self.tasks:
            #self.tasks.append(task)
        if not task.text:
            self.deleteTask(task)

    def setModified(self, val):
        self.modified = val
        self.fileModified.emit(val)

    def deleteTask(self, task):
        self.tasks.remove(task)
        self.setModified(True)

    def addTask(self, task):
        self.tasks.append(task)
        task.modified.connect(self._taskModified)
        self.setModified(True)

    def connectTask(self, task):
        task.modified.connect(self._taskModified)

    def save(self, filename=''):
        logger.debug('File.save called with filename="%s"', filename)
        self._fileObserver.clear()
        if not filename and not self.filename:
            self.filename = self._createNewFilename()
        elif filename:
            self.filename = filename
        self._saveTasks()
        self.modified = False
        self.fileModified.emit(False)
        self._fileObserver.addPath(self.filename)

    @staticmethod
    def _createNewFilename():
        newFileName = os.path.expanduser('~/todo.txt')
        if not os.path.isfile(newFileName):
            return newFileName
        for counter in range(0, 100):
            newFileName = os.path.expanduser('~/todo.{}.txt.'.format(counter))
            if not os.path.isfile(newFileName):
                return newFileName
        return os.path.expanduser('~/todo.0.txt')

    def _saveTasks(self):
        with open(self.filename, 'wt', encoding='utf-8') as fd:
            fd.writelines([(task.text + self.newline) for task in self.tasks])
        logger.debug('%s was saved to disk.', self.filename)

    def saveDoneTask(self, task):
        doneFilename = os.path.join(os.path.dirname(self.filename), 'done.txt')
        with open(doneFilename, 'at', encoding='utf-8') as fd:
            fd.write(task.text + self.newline)
        logger.debug('"%s" was appended to "%s"', task.text, doneFilename)

    def getAllContexts(self):
        return self._getAllX("contexts")

    def getAllProjects(self):
        return self._getAllX("projects")

    def _getAllX(self, name):
        res = {}
        for task in self.tasks:
            for element in getattr(task, name):
                if element not in res:
                    res[element] = [0, 0]
                idx = 1 if task.is_complete else 0
                res[element][idx] += 1
        return res

    def getAllPriorities(self):
        return self._getAllX("priority")

    def getAllDueRanges(self):
        dueRanges = dict()
        filters = [DueTodayFilter(), DueTomorrowFilter(), DueThisWeekFilter(), DueThisMonthFilter(), DueOverdueFilter()]
        for task in self.tasks:
            idx = 1 if task.is_complete else 0
            for flt in filters:
                if flt.isMatch(task):
                    if not (flt in dueRanges):
                        dueRanges[flt] = [0, 0]
                    dueRanges[flt][idx] += 1
        return dueRanges

    def getTasksCounters(self):
        counters = dict({
            'All': [0, 0],
            'Uncategorized': [0, 0],
            'Contexts': [0, 0],
            'Projects': [0, 0],
            'Complete': [0, 0],
            'Priority': [0, 0],
            'Due': [0, 0]
        })
        for task in self.tasks:
            nbProjects = len(task.projects)
            nbContexts = len(task.contexts)
            idx = 1 if task.is_complete else 0
            counters['All'][idx] += 1
            if nbProjects > 0:
                counters['Projects'][idx] += 1
            if nbContexts > 0:
                counters['Contexts'][idx] += 1
            if nbContexts == 0 and nbProjects == 0:
                counters['Uncategorized'][idx] += 1
            if task.due:
                counters['Due'][idx] += 1
            if task.priority != "":
                counters['Priority'][idx] += 1
        return counters


class FileObserver(QtCore.QFileSystemWatcher):

    fileChangetSig = QtCore.pyqtSignal(str)
    dirChangetSig = QtCore.pyqtSignal(str)

    def __init__(self):
        logger.debug('Setting up FileObserver instance.')
        super().__init__()
        self.fileChanged.connect(self.fileChangedHandler)
        self.directoryChanged.connect(self.dirChangedHandler)

    def fileChangedHandler(self, path):
        logger.debug('Detected external file change for file %s', path)
        self.removePath(path)
        self.fileChangetSig.emit(path)

    def dirChangedHandler(self, path):
        logger.debug('Detected directory change for file %s', path)
        self.dirChangetSig.emit(path)

    def clear(self):
        if self.files():
            logger.debug('Clearing watchlist.')
            self.removePaths(self.files())



import functools
import logging
import sys


def logger_name(filename, lineno):
    return filename[:-3].split('/QTodoTxt/')[1].replace('/', '.') + ':' + str(lineno)


def deprecated(func):
    '''This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being logged
    when the function is used.'''

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        func_def_line = func.__code__.co_firstlineno
        func_def_filename = func.__code__.co_filename
        prev_frame = sys._getframe().f_back
        calling_lineno = prev_frame.f_lineno
        calling_filename = prev_frame.f_code.co_filename
        logger = logging.getLogger(logger_name(calling_filename, calling_lineno))
        logger.warning("Call to deprecated function {} ({}@{}).".format
                       (func.__name__, func_def_line, func_def_filename))
        return func(*args, **kwargs)
    return new_func



from datetime import datetime
from PyQt5 import QtCore


class TaskHtmlizer(object):
    def __init__(self):
        colorSchemName = QtCore.QSettings().value("color_schem", "")
        if str(colorSchemName).find("dark") >= 0:  # for dark theme
            self.priority_colors = dict(A='red', B='#1C7F61', C='#7397BE')
            self.contextColor = "#5ED2B8"
            self.projectColor = "#FFCA73"
            self.priorityDuecolors = ['red', '#E0A180']
            self.priorityThresholdColors = ['orange', 'grey']
            self.errorColor = "red"
            self.linkColor = "#E0A180"
        else:  # for light theme
            self.priority_colors = dict(A='red', B='green', C='navy')
            self.contextColor = "green"
            self.projectColor = "#64AAD0"
            self.priorityDuecolors = ['red', 'orange']
            self.priorityThresholdColors = ['orange', 'grey']
            self.errorColor = "red"
            self.linkColor = "none"

        self.complColor = "gray"

    def task2html(self, task):
        words = task.description.split(" ")
        newwords = []
        for word in words:
            if word.startswith("@"):
                word = self._htmlizeContext(word)
            elif word.startswith("+"):
                word = self._htmlizeProject(word)
            elif word.startswith("due:"):
                word = self._htmlizeDueDate(task, word)
            elif word.startswith("t:"):
                word = self._htmlizeThresholdDate(task, word)
            elif "://" in word:
                word = self._addUrl(word, self.linkColor)
            newwords.append(word)
        html = " ".join(newwords)
        if task.is_complete:
            html = '<s>{}</s>'.format(html)
        if task.priority:
            html = self._htmlizePriority(task.priority) + html
        else:
            # add space, so tasks get evenly aligned when there's no priority
            html = '<tt>&nbsp;&nbsp;&nbsp;</tt>&nbsp;' + html
        if task.completion_date:
            html += ' <font color="{1!s}">(completed: {0!s})</font>'.format(task.completion_date, self.complColor)
        if task.creation_date:
            html += ' <font color="{1!s}">(created: {0!s})</font>'.format(task.creation_date, self.complColor)
        return html

    def _addUrl(self, word, color="none"):
        if ("http" in word) or ("www." in word):
            parts = word.split('//', 1)
            parts2 = parts[1].split('/')
            end = parts2[0]
            if len(parts2) > 1:
                end += "/..."
            cleanWord = parts[0] + '//' + end
        else:
            parts = word.split('/')
            if len(parts[-1]):
                cleanWord = parts[-1]
            else:
                cleanWord = parts[-2]
            cleanWord = "../" + cleanWord
        word = '<a style="color:{2!s};" href="{0!s}">{1!s}</a>'.format(word, cleanWord, color)
        return word

    def _htmlizeContext(self, context):
        context = context.replace("@", "")
        if "://" in context:
            context = self._addUrl(context, self.contextColor)

        return '<font style="color:{0!s}">@{1!s}</font>'.format(self.contextColor, context)

    def _htmlizeProject(self, project):
        project = project.replace("+", "")
        if "://" in project:
            project = self._addUrl(project, self.projectColor)

        return '<font style="color:{0!s}">+{1!s}</font>'.format(self.projectColor, project)

    def _htmlizePriority(self, priority):
        if priority in self.priority_colors:
            color = self.priority_colors[priority]
            return '<font style="color:{}"><tt>({})</tt>&nbsp;</font>'.format(color, priority)
        return '<tt>(%s)</tt>&nbsp;' % priority

    def _htmlizeDueDate(self, task, string):
        if not task.due:
            return ('<b><font style="color:{1!s}">*** {0!s} Invalid date format,'
                    ' expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>'.format(string, self.errorColor))

        date_now = datetime.today()
        tdelta = task.due - date_now
        if tdelta.days > 7:
            return '<b>due:{}</b>'.format(task.dueString)
        elif tdelta.days > 0:
            return '<b><font style="color:{1!s}">due:{0!s}</font></b>'.format(task.dueString, self.priorityDuecolors[1])
        else:
            return '<b><font style="color:{1!s}">due:{0!s}</font></b>'.format(task.dueString, self.priorityDuecolors[0])

    def _htmlizeThresholdDate(self, task, string):
        if not task.threshold:
            return ('<b><font style="color:{1!s}">*** {0!s} Invalid date format,'
                    ' expected yyyy-mm-dd or yyyy-mm-ddThh:mm. ***</font></b>'.format(string, self.errorColor))

        date_now = datetime.today()
        tdelta = task.threshold - date_now
        if tdelta.days > 0:
            return '<i><font style="color:{1!s}">t:{0!s}</font></i>'.\
                format(task.thresholdString, self.priorityThresholdColors[1])
        else:
            return '<font style="color:{1!s}">t:{0!s}</font>'.\
                format(task.thresholdString, self.priorityThresholdColors[0])



import unittest
from PyQt5 import QtCore
from datetime import date, timedelta

from qtodotxt2.lib import tasklib
from qtodotxt2.lib.file import File
from qtodotxt2.lib.filters import IncompleteTasksFilter, ContextFilter, ProjectFilter, DueThisMonthFilter, \
    DueThisWeekFilter, DueTodayFilter, DueOverdueFilter
from qtodotxt2.main_controller import MainController


class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        QtCore.QCoreApplication.setOrganizationName("QTodoTxt")
        QtCore.QCoreApplication.setApplicationName("Testing")
        QtCore.QSettings().setValue("Preferences/auto_save", False)
        cls.ctrl = MainController([])
        cls.ctrl.allTasks = cls.make_tasks()
        cls.ctrl.showCompleted = True
    
    @staticmethod
    def make_tasks():
        today = date.today().isoformat()
        tomorrow = (date.today() + timedelta(days=1)).isoformat()
        nextweek = (date.today() + timedelta(days=8)).isoformat()
        tasks = []
        t = tasklib.Task("(A) Task home due:{} +project1 @context2".format(today))
        tasks.append(t)
        t = tasklib.Task("(B) Task due:{} +project2 @context1".format(tomorrow))
        tasks.append(t)
        t = tasklib.Task("Task due:2015-04-01 +project2 @context1")
        tasks.append(t)
        t = tasklib.Task("TOTO due:2015-04-02 +project3")
        tasks.append(t)
        t = tasklib.Task("TOTO due:{}".format(tomorrow))
        tasks.append(t)
        t = tasklib.Task("x (B) Task due:{} +project1 @context1".format(tomorrow))
        tasks.append(t)
        t = tasklib.Task("x (B) Task due:{} +project2 @context3".format(tomorrow))
        tasks.append(t)
        t = tasklib.Task("(B) Task home +project2 @context3")
        tasks.append(t)
        return tasks


    @classmethod
    def tearDownClass(cls):
        pass

    def test_completed(self):
        self.ctrl.showCompleted = True
        self.ctrl.applyFilters()
        self.assertEqual(len(self.ctrl.filteredTasks), len(self.ctrl.allTasks))
        self.ctrl.showCompleted = False
        self.ctrl.applyFilters()
        self.assertEqual(len(self.ctrl.filteredTasks), len(self.ctrl.allTasks) - 2)
        self.ctrl.showCompleted = True

    def test_new_delete(self):
        count = len(self.ctrl.allTasks)
        idx = self.ctrl.newTask("My funny new task + PeaceProject")
        self.assertEqual(count + 1,  len(self.ctrl.allTasks))
        task = self.ctrl.filteredTasks[idx]
        self.ctrl.deleteTask(task)
        self.assertEqual(count,  len(self.ctrl.allTasks))

    def test_filter(self):
        self.ctrl.applyFilters([DueTodayFilter()])
        self.assertEqual(len(self.ctrl.filteredTasks), 1)
        print("START")
        self.ctrl.applyFilters([DueOverdueFilter()])
        self.assertEqual(len(self.ctrl.filteredTasks), 2)
        self.ctrl.applyFilters([])

    def test_filter_string(self):
        self.ctrl.searchText = "home"
        self.assertEqual(len(self.ctrl.filteredTasks), 2)
        self.ctrl.searchText = "+project2"
        self.assertEqual(len(self.ctrl.filteredTasks), 4)
        self.ctrl.searchText = "!due home"
        self.assertEqual(len(self.ctrl.filteredTasks), 1)
        self.ctrl.searchText = ""

    def test_filter_or(self):
        self.ctrl.searchText = "home | TOTO"
        self.assertEqual(len(self.ctrl.filteredTasks), 4)
